/*******************************************************************************
* Misc. useful routines to handle rotation transformations of coordinate
* systems.
* 
*	Project: Monte Carlo Simulation of Triple Axis Spectrometers
*	File name: rotation.c
*
*	Author: K.N.			Aug  8, 1997
*
*	$Id: rotation.c,v 1.3 1999/01/25 11:33:05 kn Exp $
*
*	$Log: rotation.c,v $
*	Revision 1.3  1999/01/25 11:33:05  kn
*	Remove unused variable to avoid Win32 compiler warning.
*
*	Revision 1.2  1998/10/02 08:39:20  kn
*	Fixed header comment.
*
*	Revision 1.1  1997/08/13 09:16:04  kn
*	Initial revision
*
*
* Copyright (C) Risoe National Laboratory, 1997-1998, All rights reserved
*******************************************************************************/

#include <math.h>

#include "mcstas.h"

/*******************************************************************************
* The Rotation type implements a rotation transformation of a coordinate
* system in the form of a double[3][3] matrix.
*
* Contrary to the Coords type in coords.c, rotations are passed by
* reference. Functions that yield new rotations do so by writing to an
* explicit result parameter; rotations are not returned from functions. The
* reason for this is that arrays cannot by returned from functions (though
* structures can; thus an alternative would have been to wrap the
* double[3][3] array up in a struct). Such are the ways of C programming.
*
* A rotation represents the tranformation of the coordinates of a vector when
* changing between coordinate systems that are rotated with respect to each
* other. For example, suppose that coordinate system Q is rotated 45 degrees
* around the Z axis with respect to coordinate system P. Let T be the
* rotation transformation representing a 45 degree rotation around Z. Then to
* get the coordinates of a vector r in system Q, apply T to the coordinates
* of r in P. If r=(1,0,0) in P, it will be (sqrt(1/2),-sqrt(1/2),0) in
* Q. Thus we should be careful when interpreting the sign of rotation angles:
* they represent the rotation of the coordinate systems, not of the
* coordinates (which has opposite sign).
*******************************************************************************/


/* Get the unit transformation. */
void
rot_set_unit(Rotation t)
{
  t[0][0] = 1;
  t[0][1] = 0;
  t[0][2] = 0;
  t[1][0] = 0;
  t[1][1] = 1;
  t[1][2] = 0;
  t[2][0] = 0;
  t[2][1] = 0;
  t[2][2] = 1;
}

/*******************************************************************************
* Get transformation for rotation first phx around x axis, then phy around y,
* then phz around z.
*******************************************************************************/
void
rot_set_rotation(Rotation t, double phx, double phy, double phz)
{
  double cx = cos(phx);
  double sx = sin(phx);
  double cy = cos(phy);
  double sy = sin(phy);
  double cz = cos(phz);
  double sz = sin(phz);

  t[0][0] = cy*cz;
  t[0][1] = sx*sy*cz + cx*sz;
  t[0][2] = sx*sz - cx*sy*cz;
  t[1][0] = -cy*sz;
  t[1][1] = cx*cz - sx*sy*sz;
  t[1][2] = sx*cz + cx*sy*sz;
  t[2][0] = sy;
  t[2][1] = -sx*cy;
  t[2][2] = cx*cy;
}

/*******************************************************************************
* Individual rotations around the three axis.
*******************************************************************************/
void
rot_set_rotation_x(Rotation t, double ph)
{
  double c = cos(ph);
  double s = sin(ph);

  t[0][0] = 1;
  t[0][1] = 0;
  t[0][2] = 0;
  t[1][0] = 0;
  t[1][1] = c;
  t[1][2] = s;
  t[2][0] = 0;
  t[2][1] = -s;
  t[2][2] = c;
}

void rot_set_rotation_y(Rotation t, double ph)
{
  double c = cos(ph);
  double s = sin(ph);

  t[0][0] = c;
  t[0][1] = 0;
  t[0][2] = -s;
  t[1][0] = 0;
  t[1][1] = 1;
  t[1][2] = 0;
  t[2][0] = s;
  t[2][1] = 0;
  t[2][2] = c;
}

void rot_set_rotation_z(Rotation t, double ph)
{
  double c = cos(ph);
  double s = sin(ph);

  t[0][0] = c;
  t[0][1] = s;
  t[0][2] = 0;
  t[1][0] = -s;
  t[1][1] = c;
  t[1][2] = 0;
  t[2][0] = 0;
  t[2][1] = 0;
  t[2][2] = 1;
}


/*******************************************************************************
* Matrix multiplication of transformations (this corresponds to combining
* transformations). After rot_mul(T1, T2, T3), doing T3 is equal to doing
* first T2, then T1.
* Note that T3 must no alias (use the same array as) T1 or T2.
*******************************************************************************/
void
rot_mul(Rotation t1, Rotation t2, Rotation t3)
{
  int i,j;

  for(i = 0; i < 3; i++)
    for(j = 0; j < 3; j++)
      t3[i][j] = t1[i][0]*t2[0][j] + t1[i][1]*t2[1][j] + t1[i][2]*t2[2][j];
}

/*******************************************************************************
* Copy a rotation transformation (needed since arrays cannot be assigned in C).
*******************************************************************************/
void
rot_copy(Rotation dest, Rotation src)
{
  dest[0][0] = src[0][0];
  dest[0][1] = src[0][1];
  dest[0][2] = src[0][2];
  dest[1][0] = src[1][0];
  dest[1][1] = src[1][1];
  dest[1][2] = src[1][2];
  dest[2][0] = src[2][0];
  dest[2][1] = src[2][1];
  dest[2][2] = src[2][2];
}
